#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glut.h>

#define BMP           19778
#define BI_RGB        0L
#define BI_RLE8       1L
#define BI_RLE4       2L
#define BI_BITFIELDS  3L

typedef GLubyte       BYTE;
typedef GLushort      WORD;
typedef GLuint        DWORD;
typedef GLint         LONG;

typedef struct tagRGBQUAD {
        BYTE    rgbBlue;
        BYTE    rgbGreen;
        BYTE    rgbRed;
        BYTE    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER{
        DWORD      biSize;
        LONG       biWidth;
        LONG       biHeight;
        WORD       biPlanes;
        WORD       biBitCount;
        DWORD      biCompression;
        DWORD      biSizeImage;
        LONG       biXPelsPerMeter;
        LONG       biYPelsPerMeter;
        DWORD      biClrUsed;
        DWORD      biClrImportant;
} BITMAPINFOHEADER;

typedef struct tagBITMAPINFO {
    BITMAPINFOHEADER    bmiHeader;
    RGBQUAD             bmiColors[1];
} BITMAPINFO;

typedef struct tagBITMAPFILEHEADER {
        WORD    bfType;
        DWORD   bfSize;
        WORD    bfReserved1;
        WORD    bfReserved2;
        DWORD   bfOffBits;
} BITMAPFILEHEADER;

void *LoadBitmap(char *fileName, BITMAPINFO **info)
{
  FILE          *fp;        // Wskaznik pliku
  void          *bits;      // Bity bitmapy
  unsigned long bitsize,    // Rozmiar bitmapy
                infosize;   // Rozmiar nagwka
  WORD    type;
  DWORD   size;
  WORD    reserved1;
  WORD    reserved2;
  DWORD   offBits;

// Prba otwarcia pliku w trybie binarnym
  if (!(fp = fopen(fileName, "rb"))) {
    printf("Nie ma takiego pliku: \"%s\" !\n", fileName);
    return NULL;
  }

  fread(&type, sizeof(WORD), 1, fp);
  if (type != BMP)	// Sprawdzenie sygnatury BM...
  {
  // To nie jest plik bitmapy - zwr NULL...
    fclose(fp);
    return NULL;
  };

  fread(&size, sizeof(DWORD), 1, fp);
  fread(&reserved1, sizeof(WORD), 1, fp);
  fread(&reserved2, sizeof(WORD), 1, fp);
  fread(&offBits, sizeof(DWORD), 1, fp);

  infosize = offBits - 14;
  if (!(*info = (BITMAPINFO *)malloc(infosize)))
  {
  // Brak pamici na nagwek bitmapy
    fclose(fp);
    return NULL;
  };

  if (fread(*info, 1, infosize, fp) < infosize)
  {
  // Nie powidoo si odczytanie nagwka bitmapy
    free(*info);
    fclose(fp);
    return NULL;
  };

// Gdy mamy wszystkie informacje, alokowanie pamici na bitmap i jej odczyt z dysku
  if (!(bitsize = (*info)->bmiHeader.biSizeImage))
    bitsize = ((*info)->bmiHeader.biWidth * (*info)->bmiHeader.biBitCount + 7) /
               8 * abs((*info)->bmiHeader.biHeight);

  if (!(bits = malloc(bitsize)))
  {
  // Brak pamici
    free(*info);
    fclose(fp);
    return NULL;
  }

  if (fread(bits, 1, bitsize, fp) < bitsize)
  {
  // Nie da si odczyta bitmapy - zwolnij pami i zwr warto NULL
    free(*info);
    free(bits);
    fclose(fp);
    return NULL;
  };

// OK, wszystko w porzadku, zwr wskaznik do bitmapy
  fclose(fp);
  return bits;
}

GLubyte *ConvertRGB(BITMAPINFO *info, /* We - Nagwek oryginalnej bitmapy */
           void       *bits) /* We - Piksele oryginalnej bitmapy */
{
  int   i, j,
        bitsize,  /* Cakowity rozmiar bitmapy */
        width;    /* Wyrwnana szeroko bitmapy */
  GLubyte   *newbits;   /* Nowe bity RGB */
  GLubyte   *from, *to; /* Zmienne ptli RGB */


// Alokacja pamici na bitmap RGB
  width   = 3 * info->bmiHeader.biWidth;
  width   = (width + 3) & ~3;
  bitsize = width * info->bmiHeader.biHeight;
  if ((newbits = (GLubyte *)calloc(bitsize, 1)) == NULL) return (NULL);

// Skopiowanie oryginalnej bitmapy do nowej tablicy,
// w miar potrzeby dokonanie konwersji
  switch (info->bmiHeader.biCompression)
  {
    case BI_RGB :
        if (info->bmiHeader.biBitCount == 24)
        {
            for (i = 0; i < info->bmiHeader.biHeight; i ++)
                for (j = 0, from = ((GLubyte *)bits) + i * width,
                     to = newbits + i * width;
                     j < info->bmiHeader.biWidth;
                     j ++, from += 3, to += 3)
                {
                    to[0] = from[2];
                    to[1] = from[1];
                    to[2] = from[0];
                }
        }
        break;
    case BI_RLE4 :
    case BI_RLE8 :
    case BI_BITFIELDS :
        break;
  };

  return (newbits);
}
